from __future__ import annotations

from typing import TypedDict, TypeVar, Union, cast

from .ast_node import AstNode
from .errors import AstBuilderException
from .parser_types import (
    Background,
    Cell,
    DataTable,
    DocString,
    Examples,
    Feature,
    GherkinDocument,
    Rule,
    Scenario,
    Step,
    TableRow,
    Tag,
)
from .stream.id_generator import IdGenerator
from .token import Location, Token

T = TypeVar("T")


class Comment(TypedDict):
    location: Location
    text: str


class AstBuilder:
    id_generator: IdGenerator
    stack: list[AstNode]
    comments: list[Comment]
    id_counter: int

    def __init__(self, id_generator: IdGenerator | None = None) -> None:
        if id_generator is None:
            id_generator = IdGenerator()
        self.id_generator = id_generator
        self.reset()

    def reset(self) -> None:
        self.stack = [AstNode("None")]
        self.comments = []
        self.id_counter = 0

    def start_rule(self, rule_type: str) -> None:
        self.stack.append(AstNode(rule_type))

    def end_rule(self, rule_type: str) -> None:
        node = self.stack.pop()
        self.current_node.add(node.rule_type, self.transform_node(node))

    def build(self, token: Token) -> None:
        if token.matched_type == "Comment":
            self.comments.append(
                {"location": self.get_location(token), "text": token.matched_text},
            )
        else:
            self.current_node.add(token.matched_type, token)

    def get_result(self) -> object:
        return self.current_node.get_single("GherkinDocument")

    @property
    def current_node(self) -> AstNode:
        return self.stack[-1]

    @staticmethod
    def get_location(token: Token, column: int | None = None) -> Location:
        return (
            token.location
            if not column
            else {"line": token.location["line"], "column": column}
        )

    def get_tags(self, node: AstNode) -> list[Tag]:
        tags: list[Tag] = []
        tags_node = cast(AstNode, node.get_single("Tags"))
        if not tags_node:
            return tags

        for token in tags_node.get_tokens("TagLine"):
            tags += [
                {
                    "id": self.id_generator.get_next_id(),
                    "location": self.get_location(token, tag_item["column"]),
                    "name": tag_item["text"],
                }
                for tag_item in token.matched_items
            ]

        return tags

    def get_table_rows(self, node: AstNode) -> list[TableRow]:
        rows: list[TableRow] = [
            {
                "id": self.id_generator.get_next_id(),
                "location": self.get_location(token),
                "cells": self.get_cells(token),
            }
            for token in node.get_tokens("TableRow")
        ]
        self.ensure_cell_count(rows)
        return rows

    @staticmethod
    def ensure_cell_count(rows: list[TableRow]) -> None:
        if not rows:
            return

        cell_count = len(rows[0]["cells"])
        for row in rows:
            if len(row["cells"]) != cell_count:
                raise AstBuilderException(
                    "inconsistent cell count within the table",
                    row["location"],
                )

    def get_cells(self, table_row_token: Token) -> list[Cell]:
        return [
            cast(
                Cell,
                self.reject_nones(
                    {
                        "location": self.get_location(
                            table_row_token,
                            cell_item["column"],
                        ),
                        "value": cell_item["text"],
                    },
                ),
            )
            for cell_item in table_row_token.matched_items
        ]

    @staticmethod
    def get_description(node: AstNode) -> str:
        return cast(str, node.get_single("Description", ""))

    @staticmethod
    def get_steps(node: AstNode) -> list[Step]:
        return cast(list[Step], node.get_items("Step"))

    def transform_node(
        self,
        node: AstNode,
    ) -> (
        Step
        | DocString
        | DataTable
        | Background
        | Scenario
        | Examples
        | list[TableRow]
        | str
        | None
        | Rule
        | Feature
        | GherkinDocument
        | AstNode
    ):
        if node.rule_type == "Step":
            step_line = node.get_token("StepLine")
            step_argument_type = "dummy_type"
            step_argument = None
            if node.get_single("DataTable"):
                step_argument_type = "dataTable"
                step_argument = node.get_single("DataTable")
            elif node.get_single("DocString"):
                step_argument_type = "docString"
                step_argument = node.get_single("DocString")

            return self.reject_nones(
                {
                    "id": self.id_generator.get_next_id(),
                    "location": self.get_location(step_line),
                    "keyword": step_line.matched_keyword,
                    "keywordType": step_line.matched_keyword_type,
                    "text": step_line.matched_text,
                    step_argument_type: step_argument,
                },
            )
        if node.rule_type == "DocString":
            separator_token = node.get_tokens("DocStringSeparator")[0]
            media_type = (
                separator_token.matched_text
                if len(separator_token.matched_text) > 0
                else None
            )
            line_tokens = node.get_tokens("Other")
            content = "\n".join([t.matched_text for t in line_tokens])

            return self.reject_nones(
                {
                    "location": self.get_location(separator_token),
                    "content": content,
                    "delimiter": separator_token.matched_keyword,
                    "mediaType": media_type,
                },
            )
        if node.rule_type == "DataTable":
            rows = self.get_table_rows(node)
            return self.reject_nones(
                {
                    "location": rows[0]["location"],
                    "rows": rows,
                },
            )
        if node.rule_type == "Background":
            background_line = node.get_token("BackgroundLine")
            description = self.get_description(node)
            steps = self.get_steps(node)

            return self.reject_nones(
                {
                    "id": self.id_generator.get_next_id(),
                    "location": self.get_location(background_line),
                    "keyword": background_line.matched_keyword,
                    "name": background_line.matched_text,
                    "description": description,
                    "steps": steps,
                },
            )
        if node.rule_type == "ScenarioDefinition":
            tags = self.get_tags(node)
            scenario_node = cast(AstNode, node.get_single("Scenario"))
            scenario_line = scenario_node.get_token("ScenarioLine")
            description = self.get_description(scenario_node)
            steps = self.get_steps(scenario_node)
            examples = scenario_node.get_items("ExamplesDefinition")

            return self.reject_nones(
                {
                    "id": self.id_generator.get_next_id(),
                    "tags": tags,
                    "location": self.get_location(scenario_line),
                    "keyword": scenario_line.matched_keyword,
                    "name": scenario_line.matched_text,
                    "description": description,
                    "steps": steps,
                    "examples": examples,
                },
            )
        if node.rule_type == "ExamplesDefinition":
            tags = self.get_tags(node)
            examples_node = cast(AstNode, node.get_single("Examples"))
            examples_line = examples_node.get_token("ExamplesLine")
            description = self.get_description(examples_node)
            examples_table_rows = cast(
                list[TableRow],
                examples_node.get_single("ExamplesTable"),
            )
            table_header = examples_table_rows[0] if examples_table_rows else None
            table_body = examples_table_rows[1:] if examples_table_rows else []

            return self.reject_nones(
                {
                    "id": self.id_generator.get_next_id(),
                    "tags": tags,
                    "location": self.get_location(examples_line),
                    "keyword": examples_line.matched_keyword,
                    "name": examples_line.matched_text,
                    "description": description,
                    "tableHeader": table_header,
                    "tableBody": table_body,
                },
            )
        if node.rule_type == "ExamplesTable":
            return self.get_table_rows(node)
        if node.rule_type == "Description":
            line_tokens = node.get_tokens("Other")
            tokens = list(line_tokens)

            # Trim trailing empty lines
            while tokens and not tokens[-1].matched_text:
                tokens.pop()

            return "\n".join(token.matched_text for token in tokens)
        if node.rule_type == "Rule":
            header = cast(Union[AstNode, None], node.get_single("RuleHeader"))
            if not header:
                return None
            tags = self.get_tags(header)
            rule_line = header.get_token("RuleLine")
            if not rule_line:
                return None

            children = []
            background = node.get_single("Background")
            if background:
                children.append({"background": background})
            children = children + [
                {"scenario": i} for i in node.get_items("ScenarioDefinition")
            ]
            description = self.get_description(header)

            return self.reject_nones(
                {
                    "id": self.id_generator.get_next_id(),
                    "tags": tags,
                    "location": self.get_location(rule_line),
                    "keyword": rule_line.matched_keyword,
                    "name": rule_line.matched_text,
                    "description": description,
                    "children": children,
                },
            )
        if node.rule_type == "Feature":
            header = cast(Union[AstNode, None], node.get_single("FeatureHeader"))
            if not header:
                return None

            tags = self.get_tags(header)
            feature_line = header.get_token("FeatureLine")
            if not feature_line:
                return None

            children = []
            background = node.get_single("Background")
            if background:
                children.append({"background": background})
            children = children + [
                {"scenario": i} for i in node.get_items("ScenarioDefinition")
            ]
            children = children + [{"rule": i} for i in node.get_items("Rule")]
            description = self.get_description(header)
            language = feature_line.matched_gherkin_dialect

            return self.reject_nones(
                {
                    "tags": tags,
                    "location": self.get_location(feature_line),
                    "language": language,
                    "keyword": feature_line.matched_keyword,
                    "name": feature_line.matched_text,
                    "description": description,
                    "children": children,
                },
            )
        if node.rule_type == "GherkinDocument":
            feature = node.get_single("Feature")

            return self.reject_nones({"feature": feature, "comments": self.comments})
        return node

    @staticmethod
    def reject_nones(values: dict[str, T | None]) -> dict[str, T]:
        return {k: v for k, v in values.items() if v is not None}



from .gherkin_line import GherkinLine
from .parser_types import Location


class Token:
    def __init__(self, gherkin_line: GherkinLine, location: Location):
        self.line = gherkin_line
        self.location = location

    def eof(self) -> bool:
        return not self.line

    def detach(self) -> None:
        pass  # TODO: detach line - is this needed?

    def token_value(self) -> str:
        return "EOF" if self.eof() else self.line.get_line_text()



from __future__ import annotations

from collections import defaultdict
from typing import TypeVar

from .token import Token

T = TypeVar("T")


class AstNode:
    def __init__(self, rule_type: str) -> None:
        self.rule_type = rule_type
        self._sub_items: defaultdict[str, list[object]] = defaultdict(list)

    def add(self, rule_type: str | None, obj: object) -> None:
        self._sub_items[rule_type].append(obj)

    def get_single(self, rule_type: str, default_value: object = None) -> object:
        return (
            self._sub_items[rule_type][0]
            if self._sub_items[rule_type]
            else default_value
        )

    def get_items(self, rule_type: str) -> list[object]:
        return self._sub_items[rule_type]

    def get_token(self, token_type: str) -> Token:
        return self.get_single(token_type)

    def get_tokens(self, token_type: str) -> list[Token]:
        return self._sub_items[token_type]



from __future__ import annotations

import json
from pathlib import Path
from typing import TypedDict

from typing_extensions import Self

DIALECT_FILE_PATH = Path(__file__).parent / "gherkin-languages.json"

DialectSpec = TypedDict(
    "DialectSpec",
    {
        "and": list[str],
        "background": list[str],
        "but": list[str],
        "examples": list[str],
        "feature": list[str],
        "given": list[str],
        "rule": list[str],
        "scenario": list[str],
        "scenarioOutline": list[str],
        "then": list[str],
        "when": list[str],
    },
)

with DIALECT_FILE_PATH.open(encoding="utf-8") as file:
    DIALECTS: dict[str, DialectSpec] = json.load(file)


class Dialect:
    @classmethod
    def for_name(cls, name: str) -> Self | None:
        return cls(DIALECTS[name]) if name in DIALECTS else None

    def __init__(self, spec: DialectSpec) -> None:
        self.spec = spec

    @property
    def feature_keywords(self) -> list[str]:
        return self.spec["feature"]

    @property
    def rule_keywords(self) -> list[str]:
        return self.spec["rule"]

    @property
    def scenario_keywords(self) -> list[str]:
        return self.spec["scenario"]

    @property
    def scenario_outline_keywords(self) -> list[str]:
        return self.spec["scenarioOutline"]

    @property
    def background_keywords(self) -> list[str]:
        return self.spec["background"]

    @property
    def examples_keywords(self) -> list[str]:
        return self.spec["examples"]

    @property
    def given_keywords(self) -> list[str]:
        return self.spec["given"]

    @property
    def when_keywords(self) -> list[str]:
        return self.spec["when"]

    @property
    def then_keywords(self) -> list[str]:
        return self.spec["then"]

    @property
    def and_keywords(self) -> list[str]:
        return self.spec["and"]

    @property
    def but_keywords(self) -> list[str]:
        return self.spec["but"]



from __future__ import annotations

import re
from collections.abc import Iterable

from .gherkin_line import Cell
from .parser_types import Location
from .token import Token
from .token_matcher import MatchedItems, TokenMatcher

KEYWORD_PREFIX_BULLET = "^(\\s*[*+-]\\s*)"
KEYWORD_PREFIX_HEADER = "^(#{1,6}\\s)"


class GherkinInMarkdownTokenMatcher(TokenMatcher):
    def __init__(self, dialect_name: str = "en") -> None:
        super().__init__(dialect_name)

    def reset(self) -> None:
        super().reset()
        self.matched_feature_line = False

    def match_FeatureLine(self, token: Token) -> bool:
        if self.matched_feature_line:
            return False

        # We first try to match "# Feature: blah"
        result = self._match_title_line(
            KEYWORD_PREFIX_HEADER,
            self.dialect.feature_keywords,
            ":",
            token,
            "FeatureLine",
        )
        # If we didn't match "# Feature: blah", we still match this line
        # as a FeatureLine.
        # The reason for this is that users may not want to be constrained by having this as their fist line.

        if not result:
            self._set_token_matched(token, "FeatureLine", token.line.get_line_text())
        self.matched_feature_line = True
        return True

    def match_RuleLine(self, token: Token) -> bool:
        return self._match_title_line(
            KEYWORD_PREFIX_HEADER,
            self.dialect.rule_keywords,
            ":",
            token,
            "RuleLine",
        )

    def match_ScenarioLine(self, token: Token) -> bool:
        return self._match_title_line(
            KEYWORD_PREFIX_HEADER,
            self.dialect.scenario_keywords,
            ":",
            token,
            "ScenarioLine",
        ) or self._match_title_line(
            KEYWORD_PREFIX_HEADER,
            self.dialect.scenario_outline_keywords,
            ":",
            token,
            "ScenarioLine",
        )

    def match_BackgroundLine(self, token: Token) -> bool:
        return self._match_title_line(
            KEYWORD_PREFIX_HEADER,
            self.dialect.background_keywords,
            ":",
            token,
            "BackgroundLine",
        )

    def match_ExamplesLine(self, token: Token) -> bool:
        return self._match_title_line(
            KEYWORD_PREFIX_HEADER,
            self.dialect.examples_keywords,
            ":",
            token,
            "ExamplesLine",
        )

    def match_TableRow(self, token: Token) -> bool:
        # Gherkin tables must be indented 2-5 spaces to be distinguishable from non-Gherkin tables

        if re.match("^\\s\\s\\s?\\s?\\s?\\|", token.line.get_line_text(0)):
            table_cells = token.line.table_cells
            if self._is_gfm_table_separator(table_cells):
                return False

            self._set_token_matched(
                token,
                "TableRow",
                keyword="|",
                items=token.line.table_cells,
            )

            return True
        return False

    def _is_gfm_table_separator(self, table_cells: list[Cell]) -> bool:
        text_of_table_cells = (x["text"] for x in table_cells)
        separator_values = [
            separator
            for separator in text_of_table_cells
            if re.match("^:?-+:?$", separator)
        ]
        return len(separator_values) > 0

    def match_StepLine(self, token: Token) -> bool:
        return self._match_title_line(
            KEYWORD_PREFIX_BULLET,
            self._sorted_step_keywords,
            "",
            token,
            "StepLine",
        )

    def match_Comment(self, token: Token) -> bool:
        if token.line.startswith("|"):
            table_cells = token.line.table_cells
            if self._is_gfm_table_separator(table_cells):
                self._set_token_matched(token, "Empty", indent=0)
                return True
        return False

    def match_Empty(self, token: Token) -> bool:
        result = False
        if token.line.is_empty():
            result = True
        if (
            not self.match_TagLine(token)
            and not self.match_FeatureLine(token)
            and not self.match_ScenarioLine(token)
            and not self.match_BackgroundLine(token)
            and not self.match_ExamplesLine(token)
            and not self.match_RuleLine(token)
            and not self.match_TableRow(token)
            and not self.match_Comment(token)
            and not self.match_Language(token)
            and not self.match_DocStringSeparator(token)
            and not self.match_EOF(token)
            and not self.match_StepLine(token)
        ):
            # neutered
            result = True

        if result:
            self._set_token_matched(token, "Empty", indent=0)
            return result
        return False

    # We've made a deliberate choice not to support `# language: [ISO 639-1]` headers or similar
    # in Markdown. Users should specify a language globally.
    def match_Language(self, token: Token) -> bool:
        if not token:
            raise ValueError("no token")
        return False

    def match_TagLine(self, token: Token) -> bool:
        matching_tags = re.finditer("`(@[^`]+)`", token.line.get_line_text())
        idx = 0
        tags: list[MatchedItems] = [
            {
                "column": token.line.indent + match.start(idx) + 2,
                "text": match.group(1),
            }
            for match in matching_tags
        ]

        if len(tags) == 0:
            return False

        self._set_token_matched(token, "TagLine", items=tags)
        return True

    def match_DocStringSeparator(self, token: Token) -> bool:
        if not self._active_doc_string_separator:
            # open
            return (
                self._match_DocStringSeparator(token, '"""', True)
                or self._match_DocStringSeparator(token, "````", True)
                or self._match_DocStringSeparator(token, "```", True)
            )
        # close
        return self._match_DocStringSeparator(
            token,
            self._active_doc_string_separator,
            False,
        )

    @staticmethod
    def _default_docstring_content_type() -> str:
        return ""

    def _match_title_line(
        self,
        prefix: str,
        keywords: Iterable[str],
        keywordSuffix: str,
        token: Token,
        token_type: str,
    ) -> bool:
        text = token.line.get_line_text()
        for keyword in keywords:
            match = re.search(
                f"{prefix}({re.escape(keyword)}){keywordSuffix}(.*)",
                text,
            )
            if match:
                indent = token.line.indent + len(match.group(1))
                matchedKeyword = match.group(2)
                # only set the keyword type if this is a step keyword
                if matchedKeyword in self.keyword_types:
                    matchedKeywordType = self.keyword_types[matchedKeyword][0]
                else:
                    matchedKeywordType = None
                self._set_token_matched(
                    token,
                    token_type,
                    match.group(3).strip(),
                    matchedKeyword,
                    keyword_type=matchedKeywordType,
                    indent=indent,
                )
                return True

        return False

    def _change_dialect(
        self,
        dialect_name: str,
        location: Location | None = None,
    ) -> None:
        super()._change_dialect(dialect_name, location)
        self._sorted_step_keywords = [
            key for key in self._sorted_step_keywords if key != "* "
        ]



from __future__ import annotations

import re
from collections.abc import Generator
from typing import TypedDict

from .errors import ParserException
from .parser_types import Location


class Cell(TypedDict):
    column: int
    text: str


class GherkinLine:
    def __init__(self, line_text: str, line_number: int) -> None:
        self._line_text = line_text
        self._line_number = line_number
        self._trimmed_line_text = line_text.lstrip()
        self.indent = len(line_text) - len(self._trimmed_line_text)

    def get_rest_trimmed(self, length: int) -> str:
        return self._trimmed_line_text[length:].strip()

    def get_line_text(self, indent_to_remove: int = -1) -> str:
        if indent_to_remove < 0 or indent_to_remove > self.indent:
            return self._trimmed_line_text
        return self._line_text[indent_to_remove:]

    def is_empty(self) -> bool:
        return not self._trimmed_line_text

    def startswith(self, prefix: str) -> bool:
        return self._trimmed_line_text.startswith(prefix)

    def startswith_title_keyword(self, keyword: str) -> bool:
        return self._trimmed_line_text.startswith(keyword + ":")

    @property
    def table_cells(self) -> list[Cell]:
        cells: list[Cell] = []
        for cell, col in self.split_table_cells(self._trimmed_line_text.strip()):
            lstripped_cell = re.sub(r"^[^\S\n]*", "", cell, flags=re.UNICODE)
            cell_indent = len(cell) - len(lstripped_cell)
            cells.append(
                {
                    "column": col + self.indent + cell_indent,
                    "text": re.sub(r"[^\S\n]*$", "", lstripped_cell, flags=re.UNICODE),
                },
            )
        return cells

    @staticmethod
    def split_table_cells(row: str) -> Generator[tuple[str, int]]:
        """
        An iterator returning all the table cells in a row with their positions,
        accounting for escaping.
        """

        row_iter = iter(row)
        col = 0
        start_col = col + 1
        cell = ""
        first_cell = True
        while True:
            char = next(row_iter, None)
            col += 1
            if char == "|":
                if first_cell:
                    # First cell (content before the first |) is skipped
                    first_cell = False
                else:
                    yield cell, start_col
                cell = ""
                start_col = col + 1
            elif char == "\\":
                char = next(row_iter, "")
                col += 1
                if char == "n":
                    cell += "\n"
                else:
                    if char not in ["|", "\\"]:
                        cell += "\\"
                    cell += char
            elif char:
                cell += char
            else:
                break
        # Last cell (content after the last |) is skipped

    @property
    def tags(self) -> list[Cell]:
        column = self.indent + 1
        uncommented_line = re.split(
            r"\s#",
            self._trimmed_line_text.strip(),
            maxsplit=2,
        )[0]
        items = uncommented_line.strip().split("@")
        tags: list[Cell] = []
        for item in items[1:]:
            tag_value = f"@{item.strip()}"
            if re.search(r"[^\S+]", tag_value) is not None:
                location: Location = {"line": self._line_number, "column": column}
                raise ParserException("A tag may not contain whitespace", location)

            tags.append({"column": column, "text": tag_value})
            column += len(item) + 1
        return tags



from __future__ import annotations

import contextlib
import io

from .gherkin_line import GherkinLine
from .parser_types import Location
from .token import Token


class TokenScanner:
    """
    The scanner reads a gherkin doc (typically read from a `.feature` file) and creates a token for
    each line.

    The tokens are passed to the parser, which outputs an AST (Abstract Syntax Tree).

    If the scanner sees a `#` language header, it will reconfigure itself dynamically to look for
    Gherkin keywords for the associated language. The keywords are defined in
    :file:`gherkin-languages.json`.
    """

    def __init__(self, source: str) -> None:
        self.io = io.StringIO(source)
        self.line_number = 0

    def read(self) -> Token:
        self.line_number += 1
        location: Location = {"line": self.line_number}
        line = self.io.readline()
        return Token((GherkinLine(line, self.line_number) if line else line), location)

    def __del__(self) -> None:
        # close file descriptor if it's still open
        with contextlib.suppress(AttributeError):
            self.io.close()
